from pathlib import Path

import dask.array as da
import numpy as np
import yaml
from dask import delayed


@delayed
def _read_dax(daxname, image_width, image_height, frame_number):
    startByte = frame_number * image_height * image_width * 2
    endByte = startByte + 2 * (image_height * image_width)

    fh = open(daxname, 'rb')
    fh.seek(startByte)
    data = fh.read(endByte - startByte)
    img = np.frombuffer(data, dtype='int16')
    img = img.reshape((image_height, image_width))
    return img


def read_dax(filename: Path) -> da.Array:
    """Read a DAX file

    Parameters
    ----------
    filename
        Full path to file name

    Returns
    -------
    data array
    """
    infname = f'{filename}'.replace('.dax', '.inf')
    conf = yaml.safe_load(open(infname).read().replace('=', ':'))
    image_width = conf['x_end']
    image_height = conf['y_end']
    images = []
    for i in range(conf['number of frames']):
        im = _read_dax(filename, image_width, image_height, i)
        im = da.from_delayed(im, shape=(image_width, image_height), dtype='uint16')
        images.append(im)
    return da.stack(images)



"""The image module contains functions to read TIFF files and to operate on them.
"""
import warnings

from imaxt_image.display.scaling import percentile, zscale  # noqa
from imaxt_image.io.tiffimage import TiffImage  # noqa

warnings.warn(
    'The image submodule is deprecated and will be removed in future versions. '
    'Please use the imaxt_image.display and imaxt_image.io modules instead!'
)



import numpy as np
import pytest
import scipy.misc
from scipy.ndimage import shift

from imaxt_image.registration.minimization import find_overlap_conf


def test_find_overlap_conf():
    im = scipy.misc.face(gray=True)
    im2 = shift(im, (100, 100))
    conf = np.ones_like(im)

    res = find_overlap_conf(im, im2, conf, conf, desp=(-100, -100))
    assert round(res.x) == -100
    assert round(res.y) == -100

    res = find_overlap_conf(im, im2, conf, conf, desp=(-90, -90))
    assert round(res.x) == -100
    assert round(res.y) == -100

    res = find_overlap_conf(im, im, conf, conf, desp=(0, 0))
    assert round(res.x) == 0
    assert round(res.y) == 0



from pathlib import Path

import numpy as np
import pytest

from imaxt_image.display.scaling import percentile, zscale
from imaxt_image.io.tiffimage import TiffImage


def imread(img):
    return np.zeros((4, 4, 128, 128))


def test_zscale():
    im = imread(None)
    vmin, vmax = zscale(im)
    assert vmin == 0.0
    assert vmax == 0.0


def test_percentile():
    im = imread(None)
    vmin, vmax = percentile(im, 50)
    assert vmin == 0.0
    assert vmax == 0.0


def test_image():
    im = TiffImage('tests/16bit.s.tif')
    assert im.shape == (10, 10)

    arr = im.asarray()
    assert arr.shape == (10, 10)


def test_image_notfound():
    with pytest.raises(FileNotFoundError):
        TiffImage('tests/notfound.tif')


def test_image_pathlib():
    im = TiffImage(Path('tests/16bit.s.tif'))
    assert im.shape == (10, 10)


def test_metadata():
    im = TiffImage(Path('tests/16bit.s.tif'))
    assert im.metadata is None


def test_ome_tiff():
    im = TiffImage(Path('tests/tubhiswt_C1.ome.tif'))
    assert 'OME' in im.metadata.as_dict()
    assert 'OME' in im.metadata.description


def test_dask():
    im = TiffImage(Path('tests/16bit.s.tif'))
    assert im.ndim == 2
    assert im.dtype == np.dtype('int16')

    da = im.to_dask()
    assert da.shape == (10, 10)

    xarr = im.to_xarray()
    assert len(xarr) == 10



import pytest
import scipy.misc
from scipy.ndimage import shift

from imaxt_image.registration.cross_correlation import find_shift


def test_find_shift():
    im = scipy.misc.face(gray=True)
    im2 = shift(im, (100, 100))

    res = find_shift(im, im2, overlap=(0.6, 0.8))
    assert res['x'] == -100
    assert res['y'] == -100
    assert res['overlap'] > 0.7

    assert res.x == res['x']
    assert 'x' in dir(res)

    with pytest.raises(AttributeError):
        res.z

    res = find_shift(im, im, overlap=(0.6, 1.0), initial_shift=(0, 0))
    assert res['x'] == 0
    assert res['y'] == 0
    assert res['overlap'] > 0.7

    res = find_shift(im, im2, overlap=(0.6, 1.0), initial_shift=(-100, -100))
    assert res['x'] == -100
    assert res['y'] == -100
    assert res['overlap'] > 0.7

    res = find_shift(im, im2, overlap=None)
    assert res['x'] == -100
    assert res['y'] == -100
    assert res['overlap'] > 0.7



from typing import Tuple

import numpy as np
from scipy.ndimage import shift


class ShiftResult(dict):
    """Represents the shift result.

    Attributes
    ----------
    x : float
        x offset
    y : float
        y offset
    overlap: float
        percentage of overlap
    error : float
        error
    iqr : float
        information quality ratio
    mi : float
        mutual information
    avg_flux : float
        average flux over overlapping area

    Notes
    -----
    Not all algorithms return all variables.
    """

    __slots__ = ['x', 'y', 'overlap', 'error', 'iqr', 'mi', 'avg_flux']

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __repr__(self):  # pragma: nocover
        if self.keys():
            m = max(map(len, list(self.keys()))) + 1
            return '\n'.join(
                [k.rjust(m) + ': ' + repr(v) for k, v in sorted(self.items())]
            )
        else:
            return self.__class__.__name__ + '()'

    def __dir__(self):
        return list(self.keys())


def extract_overlap(
    im1: np.ndarray,
    im2: np.ndarray,
    offset: Tuple[float],
    conf1: np.ndarray = None,
    conf2: np.ndarray = None,
) -> Tuple[np.ndarray]:
    """Extract overlapping area from two images based on an offset

    Parameters
    ----------
    im1
        Reference image
    im2
        Target image
    offset
        Offset in pixels between reference and target image
    conf1
        Confidence map for reference image
    conf2
        Confidence map for second image

    Returns
    -------
    image subsets
    """
    if offset == (0, 0):
        if (conf1 is not None) and (conf2 is not None):
            res = (im1, im2, conf1, conf2)
        else:
            res = (im1, im2)
        return res

    c = shift(
        np.ones_like(im1), (offset[0], offset[1]), mode='constant', cval=0, order=1
    )
    mask = c > 0
    nrows = sum(c.sum(axis=1) > 0)
    im1_out = im1[mask].reshape(nrows, -1).astype('float32')
    if conf1 is not None:
        conf1_out = conf1[mask].reshape(nrows, -1)

    c = shift(
        np.ones_like(im2), (-offset[0], -offset[1]), mode='constant', cval=0, order=1
    )
    mask = c > 0
    nrows = sum(c.sum(axis=1) > 0)
    im2_out = im2[mask].reshape(nrows, -1).astype('float32')
    if conf2 is not None:
        conf2_out = conf2[mask].reshape(nrows, -1)

    if (conf1 is not None) and (conf2 is not None):
        res = (im1_out, im2_out, conf1_out, conf2_out)
    else:
        res = (im1_out, im2_out)
    return res



from typing import Tuple

import numpy as np
from astropy.visualization import PercentileInterval, ZScaleInterval


def zscale(img: np.ndarray) -> Tuple[float, float]:
    """Determine zscale range.

    Parameters
    ----------
    img
        image array

    Returns
    -------
    zscale range
    """
    vmin, vmax = ZScaleInterval(krej=10).get_limits(img.ravel())
    return vmin, vmax


def percentile(img: np.ndarray, percentile: int) -> Tuple[float, float]:
    """Determine percentile range.

    Calculates the range (vmin, vmax) so that a percentile
    of the pixels is within those values.

    Parameters
    ----------
    img
        image array
    percentile
        Percentile value

    Returns
    -------
    percentile range
    """
    p = PercentileInterval(percentile)
    vmin, vmax = p.get_limits(img.ravel())
    return vmin, vmax



import numpy as np
from scipy.ndimage import shift


def mutual_information(
    im1: np.ndarray, im2: np.ndarray, offset: np.ndarray = None
) -> float:
    r"""Compute the mutual information between two images.

    The mutual information can be defined as:

    .. math::

        I(X; Y) = \sum_{y \in Y} \sum_{x \in X} \ p(x, y) \
        \log \left( \frac{p(x, y)}{p(x) \ p(y)} \right)

    where :math:`p(x,y)` is the joint probability function of X and
    Y and :math:`p(x)` and :math:`p(y)` are the marginal probability
    distribution functions of X and Y respectively.

    Parameters
    ----------
    im1
        reference image
    im2
        target image
    offset
        optional offset in pixels to apply to the target image

    Returns
    -------
    mutual information (MI) metric
    """
    x = offset if offset is not None else (0, 0)
    im2c = shift(im2, x, order=1, prefilter=False, mode='constant')
    mask = im2c > 0
    hgram, *_ = np.histogram2d(im1[mask], im2c[mask], bins=20)
    # Convert bins counts to probability values
    pxy = hgram / float(np.sum(hgram))
    px = np.sum(pxy, axis=1)  # marginal for x over y
    py = np.sum(pxy, axis=0)  # marginal for y over x
    px_py = px[:, None] * py[None, :]  # Broadcast to multiply marginals
    # Now we can do the calculation using the pxy, px_py 2D arrays
    nzs = pxy > 0  # Only non-zero pxy values contribute to the sum
    mi = np.sum(pxy[nzs] * np.log(pxy[nzs] / px_py[nzs]))
    return mi


def iqr(im1: np.ndarray, im2: np.ndarray, offset: np.ndarray = None) -> float:
    r"""Compute the information quality ratio betwwen two images.

    The Information Quality Ratio is defined as:

    .. math::

        IQR(X, Y) = \frac{\sum_{y \in Y} \sum_{x \in X} \ p(x, y) \log (p(x) \ p(y))}
                         {\sum_{y \in Y} \sum_{x \in X} \ p(x, y) \log p(x, y)} - 1

    where :math:`p(x,y)` is the joint probability function of X and
    Y and :math:`p(x)` and :math:`p(y)` are the marginal probability
    distribution functions of X and Y respectively.

    Parameters
    ----------
    im1
        reference image
    im2
        target image
    offset
        optional offset in pixels to apply to the target image

    Returns
    -------
    information quality ratio (IQR) metric

    References
    ----------
    Wijiya et al. , "Information Quality Ratio as a novel metric
    for mother wavelet selection", https://doi.org/10.1016/j.chemolab.2016.11.012

    """
    x = offset if offset is not None else (0, 0)
    im2c = shift(im2, x, order=1, prefilter=False, mode='constant')
    mask = im2c > 0
    hgram, *_ = np.histogram2d(im1[mask], im2c[mask], bins=20)
    # Convert bins counts to probability values
    pxy = hgram / float(np.sum(hgram))
    px = np.sum(pxy, axis=1)  # marginal for x over y
    py = np.sum(pxy, axis=0)  # marginal for y over x
    px_py = px[:, None] * py[None, :]  # Broadcast to multiply marginals
    # Now we can do the calculation using the pxy, px_py 2D arrays
    nzs = pxy > 0  # Only non-zero pxy values contribute to the sum
    iqr = np.sum(pxy[nzs] * np.log(px_py[nzs]))
    iqr = iqr / np.sum(pxy[nzs] * np.log(pxy[nzs]))
    return iqr



import asyncio
import json
import os
from contextlib import suppress
from glob import glob
from math import ceil
from pathlib import Path

import dask
import s3fs
import datashader as ds
import holoviews as hv
import holoviews.operation.datashader as hd
import numpy as np
import panel as pn
import xarray as xr
from astropy.visualization import PercentileInterval
from bokeh.models import HoverTool
from bokeh.util.serialization import make_globally_unique_id
from holoviews import streams
from holoviews.plotting.links import RangeToolLink


css = """
.custom-wbox > div.bk {
    padding-right: 10px;
}
.scrollable {
    overflow: auto !important;
}
"""

js_files = {
    "jquery": "https://code.jquery.com/jquery-1.11.1.min.js",
    "goldenlayout": "https://golden-layout.com/files/latest/js/goldenlayout.min.js",
}
css_files = [
    "https://golden-layout.com/files/latest/css/goldenlayout-base.css",
    "https://golden-layout.com/files/latest/css/goldenlayout-dark-theme.css",
]

template = """
{%% extends base %%}
<!-- goes in body -->
{%% block contents %%}
{%% set context = '%s' %%}
{%% if context == 'notebook' %%}
    {%% set slicer_id = get_id() %%}
    <div id='{{slicer_id}}'></div>
{%% endif %%}

<script>
var config = {
    settings: {
        hasHeaders: true,
        constrainDragToContainer: true,
        reorderEnabled: true,
        selectionEnabled: false,
        popoutWholeStack: false,
        blockedPopoutsThrowError: true,
        closePopoutsOnUnload: true,
        showPopoutIcon: false,
        showMaximiseIcon: true,
        showCloseIcon: false
    },
    content: [{
        type: 'row',
        content:[
            {
                type: 'component',
                componentName: 'view',
                componentState: { model: '{{ embed(roots.C) }}',
                                  title: 'Controls',
                                  width: 250,
                                  css_classes:['scrollable']},
                isClosable: false,
            },
            {
                type: 'column',
                content: [
                    {
                        type: 'row',
                        content:[
                            {
                                type: 'component',
                                componentName: 'view',
                                componentState: { model: '{{ embed(roots.A) }}', title: 'View', color: '#15191C'},
                                isClosable: false,
                            }
                        ]
                    }
                ]
            }
        ]
    }]
};

{%% if context == 'notebook' %%}
    var myLayout = new GoldenLayout( config, '#' + '{{slicer_id}}' );
    $('#' + '{{slicer_id}}').css({width: '100%%', height: '{{height}}', margin: '0px'})
{%% else %%}
    var myLayout = new GoldenLayout( config );
{%% endif %%}

myLayout.registerComponent('view', function( container, componentState ){
    const {width, css_classes} = componentState
    if(width)
      container.on('open', () => container.setSize(width, container.height))
    if (css_classes)
      css_classes.map((item) => container.getElement().addClass(item))
    container.setTitle(componentState.title)
    container.getElement().html(componentState.model);
    container.getElement().css( 'background-color', componentState.color );
    container.on('resize', () => window.dispatchEvent(new Event('resize')))
});


myLayout.init();
</script>
{%% endblock %%}
"""


def setup_notebook():
    pn.extension(js_files=js_files, raw_css=[css], css_files=css_files)
    hv.renderer("bokeh").theme = "dark_minimal"


def remove_bokeh_logo(plot, element):
    plot.state.toolbar.logo = None


async def async_play(dv, n1, n2, step, wait, saveas):
    if saveas is not None:
        for f in glob("tmpplot_???.png"):
            os.unlink(f)

    for i in range(n1, n2, step):
        dv.buffer = False
        dv.goto(i)
        dv.buffer = True
        if saveas is not None:
            element = dv.main.data[()]
            p = hv.render(element, backend="matplotlib")
            p.axes[0].set_axis_off()
            p.savefig(
                f"tmpplot_{i-n1:03d}.png", bbox_inches="tight", pad_inches=0, dpi=300
            )
        await asyncio.sleep(wait)
    os.system(f"ffmpeg -y -r 1 -i tmpplot_%03d.png -vcodec mpeg4 {saveas}")


def get_mc_config(alias):
    p = Path("~/.mc/config.json").expanduser()
    with open(p, "r") as fh:
        config = json.loads(fh.read())
    return config["aliases"][alias]


def get_s3_store(name, alias="imaxtgw"):
    config = get_mc_config(alias)
    s3 = s3fs.S3FileSystem(
        key=config["accessKey"],
        secret=config["secretKey"],
        client_kwargs={"endpoint_url": config["url"]},
    )
    s3store = None
    for ppath in [
        "processed",
        "processed0",
        "processed1",
        "processed2",
        "processed3",
        "processed4",
    ]:
        s3path = f"{ppath}/stpt/{name}/mos.zarr"
        if s3.exists(s3path):
            s3store = s3.get_mapper(s3path)
            break
    return s3store

class StptDataset:
    def __init__(self, sample):
        store = get_s3_store(sample)
        self.name = sample
        ds = xr.open_zarr(store).sel(type="mosaic")
        try:
            levels = ds.attrs["multiscale"]["datasets"]
        except TypeError:
            levels = json.loads(ds.attrs["multiscale"])["datasets"]
        self.ds = {
            k["level"]: xr.open_zarr(store, group=k["path"]).sel(type="mosaic")
            for k in levels
        }
        self.ds = {k: self.ds[k] * self.bscale + self.bzero for k in self.ds}
        self.scl = [k["level"] for k in levels]
        self.nlevels = len(self.scl)

    def __repr__(self):
        return self.ds[1].__repr__()

    def __len__(self):
        return len(self.ds[1])

    @property
    def channels(self):
        return list(self.ds[1].channel.values)

    @property
    def shape(self):
        xs = len(self.ds[1].x)
        ys = len(self.ds[1].y)
        return (ys, xs)

    @property
    def nz(self):
        return len(self.ds[1].z)

    @property
    def nslides(self):
        return self.nz * len(self.ds[1])

    @property
    def bscale(self):
        return self.ds[1].attrs.get("bscale", 0.001)

    @property
    def bzero(self):
        return self.ds[1].attrs.get("bzero", -10)

    def __getitem__(self, key):
        return self.ds[key]

    def compute_scale(self, x_range, y_range, res=900):
        if not x_range:
            return self.scl[-1]

        xdiff = abs(x_range[0] - x_range[1])
        scale = np.array([0, res * 1, res * 2, res * 4, res * 8, res * 16, res * 32])
        scl = scale.searchsorted(xdiff)
        scl = min(self.nlevels, scl)
        return self.scl[scl - 1]

    def display_interval(self, channel, percentile):
        zs = PercentileInterval(percentile)
        data = self.ds[8]["S001"].sel(channel=channel, z=0).values
        mask = data > 0
        return zs.get_limits(data[mask])

    def info(self):
        print(f"Channels: {self.channels}")
        print(f"Physical Slices: {len(self.ds[1])}")
        print(f"Optical Slices: {self.nz}")
        print(f"Mosaic size: {self.shape}")


class regrid(hd.regrid):
    def _process(self, element, key=None):

        # Compute coords, anges and size
        x, y = element.kdims
        coords = tuple(element.dimension_values(d, expanded=False) for d in [x, y])
        info = self._get_sampling(element, x, y)
        (x_range, y_range), (xs, ys), (width, height), (xtype, ytype) = info

        # This is how to make panning smoother
        # https://github.com/holoviz/holoviews/issues/4435
        xdiff = (x_range[1] - x_range[0]) / 2
        ydiff = (y_range[1] - y_range[0]) / 2

        x_range = (max(0, x_range[0] - xdiff), min(x_range[1] + xdiff, coords[0].max()))
        y_range = (max(0, y_range[0] - ydiff), min(y_range[1] + ydiff, coords[1].max()))

        # (xstart, xend), (ystart, yend) = (x_range, y_range)
        # xspan, yspan = (xend - xstart), (yend - ystart)
        interp = self.p.interpolation or None
        if interp == "bilinear":
            interp = "linear"

        # Compute bounds (converting datetimes)
        ((x0, x1), (y0, y1)), (xs, ys) = self._dt_transform(
            x_range, y_range, xs, ys, xtype, ytype
        )

        params = dict(bounds=(x0, y0, x1, y1))
        if width == 0 or height == 0:
            if width == 0:
                params["xdensity"] = 1
            if height == 0:
                params["ydensity"] = 1
            return element.clone((xs, ys, np.zeros((height, width))), **params)

        cvs = ds.Canvas(
            plot_width=width, plot_height=height, x_range=x_range, y_range=y_range
        )

        # Apply regridding to each value dimension
        regridded = {}
        arrays = self._get_xarrays(element, coords, xtype, ytype)
        agg_fn = self._get_aggregator(element, add_field=False)
        for vd, xarr in arrays.items():
            rarray = cvs.raster(xarr, upsample_method=interp, downsample_method=agg_fn)

            regridded[vd] = rarray
        regridded = xr.Dataset(regridded)

        return element.clone(
            regridded, datatype=["xarray"] + element.datatype, **params
        )


class StptDataViewer:
    def __init__(self, name):
        self.ds = StptDataset(name)
        self.name = self.ds.name
        self.dataset = self.ds
        self.buffer = True

    # Setup GUI

    def setup_template(self, height=600):
        self.tmpl = pn.Template(
            template=(template % "server"), nb_template=(template % "notebook")
        )
        self.tmpl.nb_template.globals["get_id"] = make_globally_unique_id
        self.tmpl.add_variable("height", f"{height}px")

    def setup_streams(self):
        self.range_xy = streams.RangeXY()
        self.pipe = streams.Pipe(data=[])
        self.pointer = streams.PointerXY()

    def setup_controller(self, channels=None):
        slider = pn.widgets.IntSlider(
            name="Slide No.", start=1, end=self.ds.nslides, value=1, width=200
        )
        slider.param.watch(self.update_slide, "value_throttled")

        self.controller = pn.WidgetBox(
            slider,
            css_classes=["widget-box", "custom-wbox"],
            sizing_mode="stretch_both",
        )

    # Play methods

    def play(self, start=None, end=None, step=1, wait=2, saveas=None):
        start = start or 1
        end = end or self.ds.nslides
        if end > self.ds.nslides:
            end = self.ds.nslides
        if end < 1:
            end = 1
        self.coro = asyncio.create_task(
            async_play(self, start, end, step, wait, saveas)
        )
        print(f"The display will now cycle from slices {start} to {end}")
        if saveas is not None:
            print(f"At the end, a movie file will ba saved as {saveas}")

    def stop(self):
        with suppress(Exception):
            self.coro.cancel()

    def goto(self, section):
        self.controller[0].value = section
        self.pipe.send({"section": section})

    # Image operations

    def get_rgb(self, n, scl):
        rlow, rhigh = self.rscale
        glow, ghigh = self.gscale
        blow, bhigh = self.bscale

        section = ceil(n / self.ds.nz)
        z = n - (section - 1) * self.ds.nz - 1

        r = self.ds[scl][f"S{section:03d}"].sel(channel=self.channels[0], z=z).data
        r = (r.clip(rlow, rhigh) - rlow) / (rhigh - rlow)

        g = self.ds[scl][f"S{section:03d}"].sel(channel=self.channels[1], z=z).data
        g = (g.clip(glow, ghigh) - glow) / (ghigh - glow)

        b = self.ds[scl][f"S{section:03d}"].sel(channel=self.channels[2], z=z).data
        b = (b.clip(blow, bhigh) - blow) / (bhigh - blow)

        return (r, g, b)

    def get_image(self, data=None, x_range=None, y_range=None, x=None, y=None):

        rlow, rhigh = self.rscale
        glow, ghigh = self.gscale
        blow, bhigh = self.bscale

        if not data:
            data = {}

        scl = self.ds.compute_scale(x_range, y_range, res=self.resolution)
        self.scl = scl

        n = data.get("section", 1)
        self.current_section = n

        r, g, b = self.get_rgb(n, scl)

        try:
            chunksize = r.chunksize[0]
            y1, y2 = (
                y_range[0] // scl,
                y_range[1] // scl,
            )
            x1, x2 = (
                x_range[0] // scl,
                x_range[1] // scl,
            )
            x1, x2, y1, y2 = (
                int(x1) - chunksize // 2,
                int(x2) + chunksize // 2,
                int(y1) - chunksize // 2,
                int(y2) + chunksize // 2,
            )
            x1 = max(0, x1)
            x2 = min(x2, self.ds.shape[1] // scl)
            y1 = max(0, y1)
            y2 = min(y2, self.ds.shape[0] // scl)
        except:  # noqa: E722
            x1 = y1 = 0
            y2 = self.ds.shape[0] // scl
            x2 = self.ds.shape[1] // scl

        ir = r[y1:y2, x1:x2]
        ig = g[y1:y2, x1:x2]
        ib = b[y1:y2, x1:x2]
        ir, ig, ib = dask.compute([ir, ig, ib])[0]

        im = hv.RGB(
            (
                range(x1 * scl, x2 * scl, scl),
                range(y1 * scl, y2 * scl, scl),
                ir,
                ig,
                ib,
            )
        )
        return im

    def get_image_zoom(self, data=None, x=None, y=None):
        rlow, rhigh = self.rscale
        glow, ghigh = self.gscale
        blow, bhigh = self.bscale

        if not data:
            data = {}

        width = 160

        n = data.get("section", 1)
        scl = 1

        r, g, b = self.get_rgb(n, scl)

        try:
            x1 = int(x) - width // 2
            y1 = int(y) - width // 2
            x2 = int(x) + width // 2
            y2 = int(y) + width // 2
        except Exception:
            x1 = y1 = 0
            x2 = y2 = width

        if x1 < 0:
            x1 = 0
            x2 = x1 + width
        if y1 < 0:
            y1 = 0
            y2 = y1 + width
        if x2 >= self.ds.shape[1]:
            x2 = self.ds.shape[1] - 1
            x1 = x2 - width
        if y2 >= self.ds.shape[1]:
            y2 = self.ds.shape[1] - 1
            y1 = y2 - width

        ir = r[y1:y2, x1:x2]
        ig = g[y1:y2, x1:x2]
        ib = b[y1:y2, x1:x2]
        ir, ig, ib = dask.compute([ir, ig, ib])[0]

        im = hv.RGB(
            (
                range(0, width),
                range(0, width),
                ir,
                ig,
                ib,
            )
        )
        return im

    def update_slide(self, event):
        self.pipe.send({"section": event.obj.value})

    def mainview(self):
        image = hv.DynamicMap(self.get_image, streams=[self.pipe, self.range_xy])
        if self.buffer:
            res = regrid(image)
        else:
            res = hd.regrid(image)
        return res

    def miniview(self):
        image = hv.DynamicMap(self.get_image, streams=[self.pipe])
        res = hd.regrid(image)
        return res

    def zoomview(self):
        image = hv.DynamicMap(self.get_image_zoom, streams=[self.pipe, self.pointer])
        return image

    # Viewer

    def view(
        self,
        *,
        channels,
        rscale=None,
        gscale=None,
        bscale=None,
        percentile=98,
        show_miniview=True,
        height=600,
        resolution=900,
    ):
        self.channels = channels
        self.resolution = resolution
        self.rscale = rscale or self.ds.display_interval(channels[0], percentile)
        self.gscale = gscale or self.ds.display_interval(channels[1], percentile)
        self.bscale = bscale or self.ds.display_interval(channels[2], percentile)

        self.setup_streams()
        self.setup_controller(channels=channels)
        self.setup_template(height=height)

        tooltips = [
            ("x", "$x{(0)}"),
            ("y", "$y{(0)}"),
        ]
        hover = HoverTool(tooltips=tooltips)
        self.main = self.mainview().opts(
            clone=True,
            responsive=True,
            hooks=[remove_bokeh_logo],
            default_tools=[hover],
            title=f"Sample: {self.name}",
        )

        boxes = hv.Rectangles([])
        self.box_stream = streams.BoxEdit(
            source=boxes,
            styles={"fill_color": ["yellow", "red", "green", "blue", "cyan"]},
        )
        boxes = boxes.opts(hv.opts.Rectangles(active_tools=[], fill_alpha=0.5))

        overlay = hd.regrid(hv.Image([]), streams=[self.pointer])

        if show_miniview:
            mini = (
                self.miniview()
                .clone(link=False)
                .opts(
                    width=200,
                    height=200,
                    xaxis=None,
                    yaxis=None,
                    default_tools=[],
                    shared_axes=False,
                    hooks=[remove_bokeh_logo],
                )
            )
            zoom = self.zoomview().opts(
                width=200,
                height=200,
                xaxis=None,
                yaxis=None,
                default_tools=[],
                shared_axes=False,
                hooks=[remove_bokeh_logo],
            )
            RangeToolLink(mini, self.main, axes=["x", "y"])
            self.tmpl.add_panel(
                "A",
                pn.Row(
                    pn.panel(self.main * overlay * boxes),
                    pn.Column(pn.panel(mini), pn.panel(zoom)),
                    width=400,
                    height=280,
                    sizing_mode="scale_both",
                ),
            )
        else:
            self.tmpl.add_panel("A", pn.Row(pn.panel(self.main)))
        self.tmpl.add_panel("C", self.controller)
        return self.tmpl



from typing import List

import numpy as np
from scipy.ndimage import gaussian_filter, shift
from scipy.optimize import minimize

from .mutual_information import mutual_information
from .utils import ShiftResult, extract_overlap


def compare_imgs(desp, im1, im2, co1, co2, do_print):
    """[summary]

    Parameters
    ----------
    desp : [type]
        [description]
    im1 : [type]
        [description]
    im2 : [type]
        [description]
    co1 : [type]
        [description]
    co2 : [type]
        [description]
    do_print : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    im2_desp = shift(im2, desp, mode='constant', cval=0, order=1)
    co2_desp = shift(co2, desp, mode='constant', cval=0, order=1)

    co_all = co2_desp * co1

    s = np.sum(co_all)
    if s == 0.0:
        return np.inf

    err = np.sum((co_all * (im2_desp - im1) ** 2) / (im1 + 0.001)) / s

    if do_print:  # pragma: nocover
        t_s = ''
        for t in desp:
            if np.abs(t) < 1e-3:
                t_s += ' {0:.1e}'.format(t)
            else:
                t_s += ' {0:.3f}'.format(t)
        print(t_s + ' {0:f}'.format(err))

    return err


def find_overlap_conf(im0, im1, conf0, conf1, desp: List[int], ftol=0.1, debug=False):
    """[summary]

    This calculates the displacementes that minimize the squared difference
    between images.

    Parameters
    ----------
    img_ref : [type]
        [description]
    conf_ref : [type]
        [description]
    img_obj : [type]
        [description]
    conf_obj : [type]
        [description]
    desp : []
        [description]

    Returns
    -------
    [type]
        [description]
    """
    im0_cut, im1_cut, conf0_cut, conf1_cut = extract_overlap(
        im0, im1, desp, conf0, conf1
    )

    im0_cut = gaussian_filter(im0_cut, sigma=3)
    im1_cut = gaussian_filter(im1_cut, sigma=3)

    res = minimize(
        compare_imgs,
        [0.0, 0.0],
        args=(im0_cut, im1_cut, conf0_cut, conf1_cut, debug),
        method='Powell',
        options={'ftol': ftol},
    )

    mi = mutual_information(im0_cut, im1_cut, res['x'])
    avg_flux = (im0_cut * conf0_cut).sum() / conf0_cut.sum()

    res = {
        'y': desp[0] + res['x'][0],
        'x': desp[1] + res['x'][1],
        'mi': mi,
        'avg_flux': avg_flux,
    }
    return ShiftResult(res)



import xml.etree.ElementTree as ET
from typing import Dict

import xmltodict


class Metadata:
    """Metadata reader.

    Parameters
    ----------
    description
        XML document

    """

    def __init__(self, description: str):
        self._description = description
        self.tree = ET.fromstring(self._description)

    def as_dict(self) -> Dict:
        """Return a dictionary representation of the XML metadata.
        """
        return xmltodict.parse(self._description)

    @property
    def description(self) -> str:
        return self._description



import numpy as np


def bytescale(data, cmin=None, cmax=None, high=255, low=0):
    """
    Byte scales an array (image).

    Byte scaling means converting the input image to uint8 dtype and scaling
    the range to ``(low, high)`` (default 0-255).
    If the input image already has dtype uint8, no scaling is done.

    Parameters
    ----------
    data : ndarray
        PIL image data array.
    cmin : scalar, optional
        Bias scaling of small values. Default is ``data.min()``.
    cmax : scalar, optional
        Bias scaling of large values. Default is ``data.max()``.
    high : scalar, optional
        Scale max value to `high`.  Default is 255.
    low : scalar, optional
        Scale min value to `low`.  Default is 0.

    Returns
    -------
    img_array : uint8 ndarray
        The byte-scaled array.

    Examples
    --------
    >>> img = array([[ 91.06794177,   3.39058326,  84.4221549 ],
                     [ 73.88003259,  80.91433048,   4.88878881],
                     [ 51.53875334,  34.45808177,  27.5873488 ]])
    >>> bytescale(img)
    array([[255,   0, 236],
           [205, 225,   4],
           [140,  90,  70]], dtype=uint8)
    >>> bytescale(img, high=200, low=100)
    array([[200, 100, 192],
           [180, 188, 102],
           [155, 135, 128]], dtype=uint8)
    >>> bytescale(img, cmin=0, cmax=255)
    array([[91,  3, 84],
           [74, 81,  5],
           [52, 34, 28]], dtype=uint8)

    """
    if data.dtype == np.uint8:
        return data

    if high < low:
        raise ValueError('`high` should be larger than `low`.')

    if cmin is None:
        cmin = data.min()
    if cmax is None:
        cmax = data.max()

    cscale = cmax - cmin
    if cscale < 0:
        raise ValueError('`cmax` should be larger than `cmin`.')
    elif cscale == 0:
        cscale = 1

    scale = float(high - low) / cscale
    bytedata = (data * 1.0 - cmin) * scale + 0.4999
    bytedata[bytedata > high] = high
    bytedata[bytedata < 0] = 0
    return np.cast[np.uint8](bytedata) + np.cast[np.uint8](low)



"""Read and write TIFF files."""

from pathlib import Path
from typing import Dict, List, Tuple, Union

import dask.array as da
import numpy as np
import xarray as xr
from dask import delayed
from distributed.protocol import dask_deserialize, dask_serialize

from imaxt_image.external.tifffile import TiffFile, TiffWriter
from imaxt_image.io.metadata import Metadata
from imaxt_image.io.omexmlmetadata import OMEXMLMetadata


@delayed
def _imread(filename):
    with TiffImage(filename) as img:
        return img.asarray()


class TiffImage:
    """Main class to operate with TIFF images.

    Parameters
    ----------
    path
        Full path to image including filename.

    Examples
    --------
    >>> from imaxt.image import TiffImage
    >>> im = TiffImage('image.tif')
    >>> im.shape
    .... (4, 4, 2048, 2048)
    >>> arr = im.asarray()
    >>> arr1 = arr[0, 0]
    """

    __slots__ = ["path", "tiff"]

    def __init__(self, path: Union[str, Path]) -> None:
        if not isinstance(path, Path):
            path = Path(path)
        self.path = path
        if self.path.exists():
            self.tiff = TiffFile(f"{path}")
        else:
            raise FileNotFoundError(f"File not found: {self.path}")

    @property
    def ndim(self) -> int:
        return len(self.shape)

    @property
    def shape(self) -> Tuple[int, ...]:
        """Return size of image.
        """
        return self.tiff.series[0].pages.shape

    @property
    def dtype(self):
        """Return type of data.
        """
        return self.tiff.series[0].pages.dtype

    def to_dask(self, chunks=None):
        """Return a dask.array representation of the data.
        """
        return da.from_delayed(_imread(self.path), shape=self.shape, dtype=self.dtype)

    def to_xarray(self, use_dask=True, **kwargs) -> xr.DataArray:
        """Return a DataArray representation of the data.

        Parameters
        ----------
        use_dask
            Use Dask to lazyload the array.

        Returns
        -------
        Xarray DataArray
        """
        if "coords" not in kwargs:
            kwargs["coords"] = tuple([range(d) for d in self.shape])

        if use_dask:
            arr = self.to_dask()
        else:
            arr = self.asarray()

        return xr.DataArray(arr, **kwargs)

    def asarray(self, out=None, maxworkers=1) -> np.ndarray:
        """Return TIFF image as numpy array.

        This functions returns the full content of the file as
        a multidimensional array.
        """
        return self.tiff.asarray(out=out, maxworkers=maxworkers)

    def __getitem__(self, item):
        return self.asarray()[item]

    @property
    def metadata(self) -> Metadata:
        """Return image metadata.

        Currently only returns ImageJ or OME-TIFF metadata.
        """
        description = self.tiff.pages[0].description
        if "OME" not in description:
            return None
        return Metadata(description)

    @property
    def omemetadata(self) -> OMEXMLMetadata:
        description = self.tiff.pages[0].description
        if "OME" not in description:
            return None
        return OMEXMLMetadata(self.tiff.pages[0].description)

    @staticmethod
    def write(img: np.ndarray, filename: Union[str, Path], compress: int = 0):
        """Write image array to filename.

        Parameters
        ----------
        img
            image array
        filename
            name of the file
        compress
            compression level
        """
        with TiffWriter(f"{filename}") as writer:
            writer.save(img, compress=compress)

    def close(self):
        self.tiff.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


@dask_serialize.register(TiffImage)
def serialize(image: TiffImage) -> Tuple[Dict, List[bytes]]:
    header = {}
    frames = [f"{image.path}".encode()]
    return header, frames


@dask_deserialize.register(TiffImage)
def deserialize(header: Dict, frames: List[bytes]) -> TiffImage:
    return TiffImage(frames[0].decode())



from pathlib import Path

import numpy as np
import xarray as xr
import xtiff
import zarr
from scipy import ndimage as ndi


def image_shift(data, offsets, block_info=None):
    if block_info is not None:
        ndim = len(data.shape)
        dd = data.squeeze()
        try:
            ndd = ndi.shift(dd, offsets)
        except Exception:
            ndd = data
        for _i in range(ndim - 2):
            ndd = ndd[np.newaxis, :]
    else:
        ndd = data
    return ndd


class STPTSection:
    def __init__(self, ds, meta):
        self.ds = ds.astype("uint16")
        self.ds.attrs = meta["attrs"]
        self.offsets = meta["offsets"]
        if self.offsets is not None:
            self.offsets = np.nan_to_num(meta["offsets"])
        self.meta = meta

    def to_tiff(self, name, dir=None):
        section = self.meta["section"]
        group = self.meta["group"]
        if dir is None:
            dir = Path(".")
        else:
            dir = Path(dir)
        out = f"{dir}/{name}_{section}_{group or 'l.1'}.ome.tiff"
        xtiff.to_tiff(self.data, out)
        print("Written", out)

    def sel(self, **kwargs):
        return STPTSection(self.ds.sel(**kwargs), self.meta)

    @property
    def data(self):
        dd = self.ds.data
        if self.offsets is not None:
            ndim = len(self.ds.data.shape)
            depth = np.abs(self.offsets).max() + 100
            depth = [0] * (ndim - 2) + [depth] * 2
            ndd = dd.map_overlap(image_shift, depth=depth, offsets=self.offsets)
        else:
            ndd = dd
        return ndd

    @property
    def values(self):
        return self.ds.values

    @property
    def attrs(self):
        return self.ds.attrs

    def __getitem__(self, item):
        yslice, xslice = item
        scale = self.meta["scale"]
        ystart, yend = yslice.start // scale, yslice.stop // scale
        xstart, xend = xslice.start // scale, xslice.stop // scale
        return STPTSection(self.ds[:, :, ystart:yend, xstart:xend], self.meta)

    def __repr__(self):
        return self.ds.__repr__()

    def _repr_html_(self):
        return self.ds._repr_html_()


class STPTDataset:
    def __init__(self, name, path, scale=1):
        self.path = Path(path)
        self.name = Path(name)
        self.scale = scale
        self.mos = self.path / self.name / "mos.zarr"

        self._read_bscale_bzero()
        self._read_dataset()

    def _read_bscale_bzero(self):
        z = zarr.open(f"{self.mos}", mode="r")
        self.bscale = z.attrs["bscale"]
        self.bzero = z.attrs["bzero"]

    def _read_metadata(self):
        ds0 = xr.open_zarr(f"{self.mos}")
        return ds0.attrs

    def _read_offsets(self):
        offsets = {
            s: [dx, dy]
            for s, dx, dy in zip(
                list(self.ds),
                self.attrs["cube_reg"]["abs_dx"],
                self.attrs["cube_reg"]["abs_dy"],
            )
        }
        return offsets

    def _read_dataset(self, clip=(0, 2 ** 16 - 1), multiplier=1000):
        if self.scale == 1:
            self.group = ""
        else:
            self.group = f"l.{self.scale}"
        self.ds = xr.open_zarr(f"{self.mos}", group=self.group)
        self.ds = self.ds.sel(type="mosaic") * self.bscale + self.bzero
        self.ds = self.ds.clip(clip[0], clip[1]) * multiplier
        if self.scale > 1:
            self.ds.attrs = self._read_metadata()
        try:
            self.offsets = self._read_offsets()
            self.offsets = [item / self.scale for item in self.offsets]
        except Exception:
            self.offsets = None

    def sel(self, scale=1, **kwargs):
        return STPTDataset(self.name, self.path, scale)

    @property
    def attrs(self):
        return self.ds.attrs

    def __len__(self):
        return len(self.ds)

    def __getitem__(self, key):
        if isinstance(key, int):
            item = list(self.ds)[key]
        else:
            item = key
        meta = {
            "name": self.name,
            "scale": self.scale,
            "group": self.group,
            "section": item,
            "attrs": self.attrs,
            "offsets": self.offsets,
        }
        return STPTSection(self.ds[item], meta)

    def __repr__(self):
        return self.ds.__repr__()

    def _repr_html_(self):
        return self.ds._repr_html_()



# #%L
# OME-FILES Python library for image IO.
# Copyright (c) 2017 University of Dundee
# %%
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# #L%

import sys
import xml.dom.minidom as minidom

TEXT_NODE = minidom.Node.TEXT_NODE


def get_text(node):
    c = node.firstChild
    if c is None or c.nodeType != TEXT_NODE:
        raise ValueError('node %r does not contain text' % (node,))
    return c.data


def get_children(node):
    return [_ for _ in node.childNodes if _.nodeType != TEXT_NODE]


class NodeWrapper(object):

    FIELDS = []

    def __setattr__(self, name, value):
        if name not in self.FIELDS:
            raise AttributeError(
                'attribute %s not allowed in %s' % (name, self.__class__.__name__)
            )
        super(NodeWrapper, self).__setattr__(name, value)

    @classmethod
    def fromnode(cls, node):
        if node.nodeName != cls.__name__:
            raise ValueError('unexpected node: %s' % node.nodeName)
        w = cls()
        for name, value in node.attributes.items():
            setattr(w, name, value)
        super(NodeWrapper, w).__setattr__('_children', get_children(node))
        return w


class Annotation(NodeWrapper):

    FIELDS = ['Annotator', 'ID', 'Namespace', 'Description', 'AnnotationRef']

    @classmethod
    def fromnode(cls, node):
        ann = super(Annotation, cls).fromnode(node)
        ann.AnnotationRef = []
        for c in ann._children:
            if c.nodeName == 'Description':
                ann.Description = get_text(c)
            if c.nodeName == 'AnnotationRef':
                ann.AnnotationRef.append(c.attributes.get('ID').value)
        return ann


class MapAnnotation(Annotation):

    FIELDS = Annotation.FIELDS + ['Value']

    @classmethod
    def fromnode(cls, node):
        ann = super(MapAnnotation, cls).fromnode(node)
        ann.Value = {}
        for c in ann._children:
            if c.nodeName == 'Value':
                for m in get_children(c):
                    try:
                        k = m.attributes.get('K').value
                    except AttributeError:
                        pass  # The "K" attribute is optional
                    else:
                        ann.Value.setdefault(k, []).append(get_text(m))
        return ann


class XMLAnnotation(Annotation):

    FIELDS = Annotation.FIELDS + ['Value']

    @classmethod
    def fromnode(cls, node):
        ann = super(XMLAnnotation, cls).fromnode(node)
        for c in ann._children:
            if c.nodeName == 'Value':
                ann.Value = c
        return ann


class OMEXMLMetadata(object):
    def __init__(self, xml_string):
        if sys.version_info < (3,):
            xml_string = xml_string.encode('utf-8')
        self.doc = minidom.parseString(xml_string)

    def get_map_annotations(self):
        return [
            MapAnnotation.fromnode(_)
            for _ in self.doc.getElementsByTagName('MapAnnotation')
        ]

    def get_xml_annotations(self):
        return [
            XMLAnnotation.fromnode(_)
            for _ in self.doc.getElementsByTagName('XMLAnnotation')
        ]



from typing import Tuple

import numpy as np
import xarray as xr
from scipy.fft import fft
from scipy.ndimage import fourier_gaussian, shift
from skimage.registration._phase_cross_correlation import (
    _compute_error,
    _compute_phasediff,
    _masked_phase_cross_correlation,
    _upsampled_dft,
)

from .mutual_information import iqr
from .utils import ShiftResult, extract_overlap


def find_shift(
    im0: np.ndarray,
    im1: np.ndarray,
    overlap: Tuple[float] = None,
    border_width: int = 0,
    upsample_factor: int = 1,
    initial_shift: Tuple[int] = None,
) -> ShiftResult:
    """Find shift between images using cross correlation.

    Parameters
    ----------
    im0
        Reference image
    im1
        Target image
    overlap
        Image overlap range to exclude possible offsets
    border_width
        Ignore maxima around this image border width
    upsample_factor
        Upsampling factor
    initial_shift
        Initial guess

    Returns
    -------
    The result represented as a ``ShiftResult`` object.

    References
    ----------
    See: http://www.sci.utah.edu/publications/SCITechReports/UUSCI-2006-020.pdf
    """

    if isinstance(im0, xr.DataArray):
        this_im0 = im0.values
    else:
        this_im0 = im0

    if isinstance(im1, xr.DataArray):
        this_im1 = im1.values
    else:
        this_im1 = im1

    if initial_shift is not None:
        im0_cut, im1_cut = extract_overlap(this_im0, this_im1, initial_shift)
    else:
        initial_shift = (0, 0)
        im0_cut, im1_cut = this_im0, this_im1

    ysize, xsize = im0.shape
    offset, error, phase = phase_cross_correlation(
        im0_cut,
        im1_cut,
        border_width=border_width,
        upsample_factor=upsample_factor,
        extra=True,
    )

    yyt = offset[0]
    xxt = offset[1]
    permutations = (
        (yyt, xxt),
        (yyt + ysize, xxt),
        (yyt, xxt + xsize),
        (yyt + ysize, xxt + xsize),
        (yyt - ysize, xxt),
        (yyt, xxt - xsize),
        (yyt - ysize, xxt - xsize),
    )
    res = [None, None, None]
    mi = 0
    for p in permutations:
        im = np.ones_like(im0)
        offset = [p[0] + initial_shift[0], p[1] + initial_shift[1]]
        pixels = shift(im, offset, mode="constant", cval=0, order=1).sum()
        if pixels > 0:
            nmi = iqr(im0_cut, im1_cut, offset=p)
            if nmi > mi:
                if overlap is not None:
                    if (
                        pixels >= xsize * ysize * overlap[0]
                        and pixels <= xsize * ysize * overlap[1]
                    ):
                        res = [offset[0], offset[1], pixels / xsize / ysize, mi]
                        mi = nmi
                else:
                    res = [offset[0], offset[1], pixels / xsize / ysize, mi]
                    mi = nmi

    res = {"y": res[0], "x": res[1], "overlap": res[2], "error": error, "iqr": mi}
    return ShiftResult(res)


def phase_cross_correlation(  # noqa: C901
    reference_image,
    moving_image,
    *,
    upsample_factor=1,
    space="real",
    return_error=True,
    reference_mask=None,
    border_width=0,
    extra=False,
    moving_mask=None,
    overlap_ratio=0.3
):
    """Efficient subpixel image translation registration by cross-correlation.

    This code gives the same precision as the FFT upsampled cross-correlation
    in a fraction of the computation time and with reduced memory requirements.
    It obtains an initial estimate of the cross-correlation peak by an FFT and
    then refines the shift estimation by upsampling the DFT only in a small
    neighborhood of that estimate by means of a matrix-multiply DFT.

    Parameters
    ----------
    reference_image : array
        Reference image.
    moving_image : array
        Image to register. Must be same dimensionality as
        ``reference_image``.
    upsample_factor : int, optional
        Upsampling factor. Images will be registered to within
        ``1 / upsample_factor`` of a pixel. For example
        ``upsample_factor == 20`` means the images will be registered
        within 1/20th of a pixel. Default is 1 (no upsampling).
        Not used if any of ``reference_mask`` or ``moving_mask`` is not None.
    space : string, one of "real" or "fourier", optional
        Defines how the algorithm interprets input data. "real" means
        data will be FFT'd to compute the correlation, while "fourier"
        data will bypass FFT of input data. Case insensitive. Not
        used if any of ``reference_mask`` or ``moving_mask`` is not
        None.
    return_error : bool, optional
        Returns error and phase difference if on, otherwise only
        shifts are returned. Has noeffect if any of ``reference_mask`` or
        ``moving_mask`` is not None. In this case only shifts is returned.
    reference_mask : ndarray
        Boolean mask for ``reference_image``. The mask should evaluate
        to ``True`` (or 1) on valid pixels. ``reference_mask`` should
        have the same shape as ``reference_image``.
    moving_mask : ndarray or None, optional
        Boolean mask for ``moving_image``. The mask should evaluate to ``True``
        (or 1) on valid pixels. ``moving_mask`` should have the same shape
        as ``moving_image``. If ``None``, ``reference_mask`` will be used.
    overlap_ratio : float, optional
        Minimum allowed overlap ratio between images. The correlation for
        translations corresponding with an overlap ratio lower than this
        threshold will be ignored. A lower `overlap_ratio` leads to smaller
        maximum translation, while a higher `overlap_ratio` leads to greater
        robustness against spurious matches due to small overlap between
        masked images. Used only if one of ``reference_mask`` or
        ``moving_mask`` is None.

    Returns
    -------
    shifts : ndarray
        Shift vector (in pixels) required to register ``moving_image``
        with ``reference_image``. Axis ordering is consistent with
        numpy (e.g. Z, Y, X)
    error : float
        Translation invariant normalized RMS error between
        ``reference_image`` and ``moving_image``.
    phasediff : float
        Global phase difference between the two images (should be
        zero if images are non-negative).

    References
    ----------
    .. Manuel Guizar-Sicairos, Samuel T. Thurman, and James R. Fienup,
           "Efficient subpixel image registration algorithms,"
           Optics Letters 33, 156-158 (2008). DOI: 10.1364/OL.33.000156
    .. James R. Fienup, "Invariant error metrics for image reconstruction"
           Optics Letters 36, 8352-8357 (1997). DOI: 10.1364/AO.36.008352
    .. Dirk Padfield. Masked Object Registration in the Fourier Domain.
           IEEE Transactions on Image Processing, vol. 21(5),
           pp. 2706-2718 (2012). DOI: 10.1109/TIP.2011.2181402
    .. D. Padfield. "Masked FFT registration". In Proc. Computer Vision and
           Pattern Recognition, pp. 2918-2925 (2010).
           DOI: 10.1109/CVPR.2010.5540032

    """
    if (reference_mask is not None) or (moving_mask is not None):
        return _masked_phase_cross_correlation(
            reference_image, moving_image, reference_mask, moving_mask, overlap_ratio
        )

    # images must be the same shape
    if reference_image.shape != moving_image.shape:
        raise ValueError("images must be same shape")

    # assume complex data is already in Fourier space
    if space.lower() == "fourier":
        src_freq = reference_image
        target_freq = moving_image
    # real data needs to be fft'd.
    elif space.lower() == "real":
        src_freq = fft.fftn(reference_image)
        target_freq = fft.fftn(moving_image)
    else:
        raise ValueError('space argument must be "real" of "fourier"')

    # Whole-pixel shift - Compute cross-correlation by an IFFT
    shape = src_freq.shape
    image_product = src_freq * target_freq.conj()
    if extra:
        norm = (
            np.sqrt(src_freq * src_freq.conj() + target_freq * target_freq.conj())
            + 1e-10
        )
        image_product = image_product / norm
        image_product = fourier_gaussian(image_product, 5)

    cross_correlation = fft.ifftn(image_product)

    if border_width > 0:
        cross_correlation[:border_width, :] = 0
        cross_correlation[-border_width:, :] = 0
        cross_correlation[:, :border_width] = 0
        cross_correlation[:, -border_width:] = 0

    # Locate maximum
    maxima = np.unravel_index(
        np.argmax(np.abs(cross_correlation)), cross_correlation.shape
    )
    midpoints = np.array([np.fix(axis_size / 2) for axis_size in shape])

    shifts = np.array(maxima, dtype=np.float64)
    shifts[shifts > midpoints] -= np.array(shape)[shifts > midpoints]

    if upsample_factor == 1:
        if return_error:
            src_amp = np.sum(np.abs(src_freq) ** 2) / src_freq.size
            target_amp = np.sum(np.abs(target_freq) ** 2) / target_freq.size
            CCmax = cross_correlation[maxima]
    # If upsampling > 1, then refine estimate with matrix multiply DFT
    else:
        # Initial shift estimate in upsampled grid
        shifts = np.round(shifts * upsample_factor) / upsample_factor
        upsampled_region_size = np.ceil(upsample_factor * 1.5)
        # Center of output array at dftshift + 1
        dftshift = np.fix(upsampled_region_size / 2.0)
        upsample_factor = np.array(upsample_factor, dtype=np.float64)
        normalization = src_freq.size * upsample_factor ** 2
        # Matrix multiply DFT around the current shift estimate
        sample_region_offset = dftshift - shifts * upsample_factor
        cross_correlation = _upsampled_dft(
            image_product.conj(),
            upsampled_region_size,
            upsample_factor,
            sample_region_offset,
        ).conj()
        cross_correlation /= normalization
        # Locate maximum and map back to original pixel grid
        maxima = np.unravel_index(
            np.argmax(np.abs(cross_correlation)), cross_correlation.shape
        )
        CCmax = cross_correlation[maxima]

        maxima = np.array(maxima, dtype=np.float64) - dftshift

        shifts = shifts + maxima / upsample_factor

        if return_error:
            src_amp = _upsampled_dft(src_freq * src_freq.conj(), 1, upsample_factor)[
                0, 0
            ]
            src_amp /= normalization
            target_amp = _upsampled_dft(
                target_freq * target_freq.conj(), 1, upsample_factor
            )[0, 0]
            target_amp /= normalization

    # If its only one row or column the shift along that dimension has no
    # effect. We set to zero.
    for dim in range(src_freq.ndim):
        if shape[dim] == 1:
            shifts[dim] = 0

    if return_error:
        return (
            shifts,
            _compute_error(CCmax, src_amp, target_amp),
            _compute_phasediff(CCmax),
        )
    else:
        return shifts



from setuptools import find_packages, setup


with open("README.rst") as readme_file:
    readme = readme_file.read()

requirements = [
    "astropy",
    "numpy",
    "scipy",
    "xmltodict",
    "imagecodecs",
    "zarr",
    "dask",
    "xarray",
    "scikit-image>=0.18",
    "xtiff",
    "tifffile>=2021.7.2",
    "s3fs>=2022.2.0",
]

extras_requirements = {"pyviz": ["matplotlib", "holoviews", "datashader"]}

setup_requirements = ["pytest-runner", "flake8"]

test_requirements = ["coverage", "pytest", "pytest-cov", "pytest-mock"]

setup(
    author="Eduardi Gonzalez Solares",
    maintainer="Eduardo Gonzalez Solares",
    maintainer_email="eglez@ast.cam.ac.uk",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3.7",
    ],
    description="IMAXT Image Utilities",
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme,
    include_package_data=True,
    keywords="imaxt",
    name="imaxt-image",
    packages=find_packages(include=["imaxt*"]),
    setup_requires=setup_requirements,
    extras_requires=extras_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/IMAXT/imaxt-image",
    version="0.17.3",
    zip_safe=False,
    python_requires=">=3.7",
)



import numpy as np
import xarray as xr
import holoviews as hv
import datashader as ds
import pandas as pd
import os

from holoviews.operation.datashader import regrid
from datashader.utils import ngjit
import cv2
import scipy


class IMCViewer:
    def __init__(
        self,
        IMC_name,
        ROI_number,
        path=None,
        use_panel_file=False,
        panel_file=None,
        panel_skip_rows=4,
    ):
        self.IMC_name = IMC_name
        self.ROI_number = ROI_number
        if path:
            self.filepath = path
        else:
            self.filepath = "/data/meds1_d/storage/raw/imc/"

        print("Loading IMC dataset")
        path_to_open = (
            os.path.join(self.filepath, self.IMC_name) + "/Q00" + str(self.ROI_number)
        )
        ds = xr.open_zarr(path_to_open)
        image = ds["Q00" + str(ROI_number)]
        image.load()
        channels = pd.DataFrame(image.meta[0]["q_channels"])
        self.data = image
        self.channels = channels
        self.channels["good"] = np.zeros(len(self.channels))

        if use_panel_file:
            panel = pd.read_excel(panel_file, skiprows=panel_skip_rows)
            for item in self.channels.metal:
                if "(" in item:
                    metal = item[:2]
                    number = item[3:6]
                    tomatch = number + metal
                    match = panel[panel["Metal"] == tomatch].Target.tolist()
                    if len(match) > 0:
                        self.channels.loc[self.channels.metal == item, "target"] = match
                        self.channels.loc[self.channels.metal == item, "good"] = 1
        else:
            self.channels["good"] = np.ones(len(self.channels))

    def prepare_data(self, ch_to_display, cropped=False):
        data = []
        mins = []
        maxs = []
        colors = []

        if cropped:
            for ch in ch_to_display:
                ch_index = self.channels[
                    self.channels["target"] == ch[0]
                ].index.tolist()[0]
                data_ch = self.data_cropped[ch_index, :, :]
                if ch[4]:
                    data_ch.data = scipy.signal.medfilt2d(data_ch.data, ch[4])
                    # data_ch.data = cv2.medianBlur(data_ch.data, ch[4])
                data.append(data_ch)
                mins.append(ch[1])
                maxs.append(ch[2])
                colors.append(ch[3])
            self.selected_data_cropped = data
            self.mins_cropped = mins
            self.maxs_cropped = maxs
            self.colors_cropped = colors
        else:
            for ch in ch_to_display:
                ch_index = self.channels[
                    self.channels["target"] == ch[0]
                ].index.tolist()[0]
                data_ch = self.data[ch_index, :, :]
                if ch[4]:
                    data_ch.data = scipy.signal.medfilt2d(data_ch.data, ch[4])
                    # data_ch.data = cv2.medianBlur(data_ch.data, ch[4])
                data.append(data_ch)
                mins.append(ch[1])
                maxs.append(ch[2])
                colors.append(ch[3])
            self.selected_data = data
            self.mins = mins
            self.maxs = maxs
            self.colors = colors

    def prepare_texts(self, ch_to_display):
        texts = []
        for n in range(len(ch_to_display)):
            tx = hv.Points((0, 0), label=ch_to_display[n][5]).opts(
                color=ch_to_display[n][3]
            )
            texts.append(tx)
            self.captions_legend = texts

    def select_color_rgb(self, colour):
        if colour == "red":
            return 1, 0, 0
        elif colour == "green":
            return 0, 1, 0
        elif colour == "blue":
            return 0, 0, 1
        elif colour == "cyan":
            return 0, 1, 1
        elif colour == "yellow":
            return 1, 1, 0
        elif colour == "magenta":
            return 1, 0, 1
        elif colour == "grey":
            return 1, 1, 1

    def colorize_channel_whitefield(self, image, factors):

        rscale, gscale, bscale = factors
        r = image / image.max()
        g = image / image.max()
        b = image / image.max()
        rinv = abs(1 - r * rscale)
        ginv = abs(1 - g * gscale)
        binv = abs(1 - b * bscale)

        rgb2 = np.ones((image.shape[0], image.shape[1], 3))
        rgb2[:, :, 0] = rinv
        rgb2[:, :, 1] = ginv
        rgb2[:, :, 2] = binv

        return rgb2

    def combine_ch_colors(self):
        channels = self.selected_data
        min_values = self.mins
        max_values = self.maxs
        colours = self.colors
        xs, ys = channels[0]["x"], channels[0]["y"]

        color_arrays = []
        for col in colours:
            arr = self.select_color_rgb(col)
            color_arrays.append(arr)

        rgb_tot = np.zeros((channels[0].shape[0], channels[0].shape[1], 3))

        for n in range(len(channels)):
            ch_data = channels[n].data
            if "p" in str(max_values[n]):
                percentile = float(max_values[n][1:])
                threshold = np.percentile(ch_data, percentile)
                print(
                    "Auto Max value for channel: " + str(n) + " is: " + str(threshold)
                )
                if threshold == 0:
                    threshold = 0.01
            else:
                threshold = max_values[n]
            if "p" in str(min_values[n]):
                percentile_min = float(min_values[n][1:])
                threshold_min = np.percentile(ch_data, percentile_min)
                print(
                    "Auto Min value for channel: "
                    + str(n)
                    + " is: "
                    + str(threshold_min)
                )
                if threshold_min == 0:
                    threshold_min = 0.01
            else:
                threshold_min = min_values[n]
            ch = ds.utils.orient_array(channels[n])
            out = norm_colorize(ch, threshold_min, threshold, color_arrays[n], rgb_tot)
            rgb_tot = out

        self.rendered_image = hv.RGB(
            (xs, ys[::-1], rgb_tot[:, :, 0], rgb_tot[:, :, 1], rgb_tot[:, :, 2])
        )
        self.rgb_tot = rgb_tot

    def combine_ch_colors_cropped(self):
        channels = self.selected_data_cropped
        min_values = self.mins_cropped
        max_values = self.maxs_cropped
        colours = self.colors_cropped
        xs, ys = channels[0]["x"], channels[0]["y"]

        color_arrays = []
        for col in colours:
            arr = self.select_color_rgb(col)
            color_arrays.append(arr)

        rgb_tot = np.zeros((channels[0].shape[0], channels[0].shape[1], 3))

        for n in range(len(channels)):
            ch_data = channels[n].data
            if "p" in str(max_values[n]):
                percentile = float(max_values[n][1:])
                threshold = np.percentile(ch_data, percentile)
                print(
                    "Auto Max value for channel: " + str(n) + " is: " + str(threshold)
                )
                if threshold == 0:
                    threshold = 0.01
            else:
                threshold = max_values[n]
            if "p" in str(min_values[n]):
                percentile_min = float(min_values[n][1:])
                threshold_min = np.percentile(ch_data, percentile_min)
                print(
                    "Auto Min value for channel: "
                    + str(n)
                    + " is: "
                    + str(threshold_min)
                )
                if threshold_min == 0:
                    threshold_min = 0.01
            else:
                threshold_min = min_values[n]
            ch = ds.utils.orient_array(channels[n])
            out = norm_colorize(ch, threshold_min, threshold, color_arrays[n], rgb_tot)
            rgb_tot = out

        self.rendered_image_cropped = hv.RGB(
            (xs, ys[::-1], rgb_tot[:, :, 0], rgb_tot[:, :, 1], rgb_tot[:, :, 2])
        )
        self.rgb_tot_cropped = rgb_tot

    def scalebar(self, x_range, y_range):
        if x_range and y_range:
            x0, x1 = x_range
            y0, y1 = y_range
            x_span = x1 - x0
            y_span = y1 - y0
            y0 = y0 + int(y_span / 20)
            y1 = y0 + int(y_span / 50)
            ytext = y1 + int(y_span / 30)
            x0 = x1 - self.scalebar_size - int(x_span / 30)
            x1 = x1 - int(x_span / 30)
            rect = hv.Rectangles((x0, y0, x1, y1)).opts(color="w")
            text = hv.Text(x0, ytext, str(self.scalebar_size) + " um").opts(
                text_color="w", text_align="left", text_font="Helvetica"
            )
            scalebar = rect * text
            return scalebar
        else:
            x0 = 0
            y0 = 0
            x1 = self.rendered_image.data["x"].size
            y1 = self.rendered_image.data["y"].size
            x_span = x1 - x0
            y_span = y1 - y0
            y0 = y0 + int(y_span / 20)
            y1 = y0 + int(y_span / 50)
            x0 = x1 - self.scalebar_size - int(x_span / 40)
            x1 = x1 - int(x_span / 30)
            ytext = y1 + int(y_span / 30)
            rect = hv.Rectangles((x0, y0, x1, y1)).opts(color="w")
            text = hv.Text(x0, ytext, str(self.scalebar_size) + " um").opts(
                text_color="w", text_align="left", text_font="Helvetica"
            )
            scalebar = rect * text
            return scalebar

    def scalebar_cropped(self, x_range, y_range):
        if x_range and y_range:
            x0, x1 = x_range
            y0, y1 = y_range
            x_span = x1 - x0
            y_span = y1 - y0
            y0 = y0 + int(y_span / 20)
            y1 = y0 + int(y_span / 50)
            ytext = y1 + int(y_span / 30)
            x0 = x1 - self.scalebar_size_cropped - int(x_span / 30)
            x1 = x1 - int(x_span / 30)
            rect = hv.Rectangles((x0, y0, x1, y1)).opts(color="w")
            text = hv.Text(x0, ytext, str(self.scalebar_size_cropped) + " um").opts(
                text_color="w", text_align="left", text_font="Helvetica"
            )
            scalebar = rect * text
            return scalebar
        else:
            x0 = self.x_range[0]
            y0 = self.y_range[0]
            x1 = self.x_range[1]
            y1 = self.y_range[1]
            x_span = x1 - x0
            y_span = y1 - y0
            y0 = y0 + int(y_span / 20)
            y1 = y0 + int(y_span / 50)
            x0 = x1 - self.scalebar_size_cropped - int(x_span / 40)
            x1 = x1 - int(x_span / 30)
            ytext = y1 + int(y_span / 30)
            rect = hv.Rectangles((x0, y0, x1, y1)).opts(color="w")
            text = hv.Text(x0, ytext, str(self.scalebar_size_cropped) + " um").opts(
                text_color="w", text_align="left", text_font="Helvetica"
            )
            scalebar = rect * text
            return scalebar

    def print_channels(self):
        print(list(self.channels.target))

    def view(
        self,
        ch_to_display,
        scalebar_size,
        rescale_factor,
        show_legend=True,
        legend_position="bottom_left",
    ):

        w = self.data.sizes["x"]
        h = self.data.sizes["y"]

        self.prepare_data(ch_to_display)
        self.combine_ch_colors()
        self.prepare_texts(ch_to_display)

        self.rangexy = hv.streams.RangeXY(source=self.rendered_image)
        self.scalebar_size = scalebar_size

        self.regridded_image = (
            regrid(self.rendered_image, streams=[self.rangexy])
            .options(framewise=True)
            .opts(width=int(w / rescale_factor), height=int(h / rescale_factor))
        )
        self.scalebar_image = self.regridded_image * hv.DynamicMap(
            self.scalebar, streams=[self.rangexy]
        )

        self.overlay_image = hv.Overlay([self.scalebar_image] + self.captions_legend)
        self.final_image = self.overlay_image.collate().opts(
            show_legend=show_legend, legend_position=legend_position
        )
        return self.final_image

    def select_zoomed_ROI(self):
        self.ROI_final = self.final_image.select(
            x=self.rangexy.x_range, y=self.rangexy.y_range
        )
        self.ROI_image = self.rendered_image.select(
            x=self.rangexy.x_range, y=self.rangexy.y_range
        )
        self.x_range = self.rangexy.x_range
        self.y_range = self.rangexy.y_range
        print("ROI Selected")
        print("X_range = " + str(self.x_range))
        print("Y_range = " + str(self.y_range))

    def crop_dataset(self):
        if self.x_range and self.y_range:
            self.data_cropped = self.data.sel(
                x=range(int(self.x_range[0]), int(self.x_range[1])),
                y=range(int(self.y_range[0]), int(self.y_range[1])),
            )
        else:
            print("run select_zoomed_roi first")

    def save_image(
        self,
        filename,
        individual=False,
        scalebar=True,
        scalebar_size=100,
        return_rgb=False,
        downsample=None,
    ):

        if individual:
            r = self.ROI_image_individual.data["R"]
            g = self.ROI_image_individual.data["G"]
            b = self.ROI_image_individual.data["B"]
        else:
            r = self.ROI_image.data["R"]
            g = self.ROI_image.data["G"]
            b = self.ROI_image.data["B"]
        r = r / r.max() * 255
        g = g / g.max() * 255
        b = b / b.max() * 255
        rgb = np.zeros((r.shape[0], r.shape[1], 3))
        rgb[:, :, 0] = b
        rgb[:, :, 1] = g
        rgb[:, :, 2] = r
        rgb = cv2.flip(rgb, 0)

        if scalebar:

            h = rgb.shape[0]
            w = rgb.shape[1]

            start = (w - scalebar_size - int(w / 20), h - 10 - int(h / 20))
            end = (w - int(w / 20), h - int(h / 20))
            start_text = (w - scalebar_size - int(w / 20), h - 10 - int(h / 20) - 10)
            text = str(scalebar_size) + " um"

            rgb2 = cv2.rectangle(rgb, start, end, (255, 255, 255), -1)
            rgb3 = cv2.putText(
                rgb2,
                text,
                start_text,
                cv2.FONT_HERSHEY_SIMPLEX,
                0.6,
                (255, 255, 255),
                2,
                cv2.LINE_AA,
            )

            if downsample:
                width = int(rgb.shape[1] * downsample)
                height = int(rgb.shape[0] * downsample)
                dim = (width, height)
                rgb4 = cv2.resize(rgb3, dim, interpolation=cv2.INTER_AREA)
                cv2.imwrite(filename + ".png", rgb4)
            else:
                cv2.imwrite(filename + ".png", rgb3)

        else:
            if downsample:
                width = int(rgb.shape[1] * downsample)
                height = int(rgb.shape[0] * downsample)
                dim = (width, height)
                rgb2 = cv2.resize(rgb, dim, interpolation=cv2.INTER_AREA)
                cv2.imwrite(filename + ".png", rgb2)
            else:
                cv2.imwrite(filename + ".png", rgb)

        if return_rgb:
            return rgb

    def print_individual_images(
        self,
        channels_to_show,
        scalebar_size,
        colour,
        show_legend=False,
        legend_position="bottom_left",
        print_histogram=False,
        hist_range=(0, 255),
        hist_colour="blue",
        path=".",
        median=None,
        print_nuclear_ch=False,
        nuclear_channel=None,
        nuclear_max="p99",
        nuclear_colour="blue",
        downsample=None,
    ):

        self.individual_images = list()

        for item in channels_to_show:
            channel = item[0]
            max_value = item[1]
            if print_nuclear_ch:
                ch_to_display = [
                    [channel, 0, max_value, colour, median, channel],
                    [
                        nuclear_channel,
                        0,
                        nuclear_max,
                        nuclear_colour,
                        median,
                        nuclear_channel,
                    ],
                ]
            else:
                ch_to_display = [[channel, 0, max_value, colour, median, channel]]

            self.prepare_data(ch_to_display)
            self.combine_ch_colors()
            self.ROI_image_individual = self.rendered_image.select(
                x=self.x_range, y=self.y_range
            )
            self.individual_images.append(self.ROI_image_individual)
            rgb = self.save_image(
                os.path.join(path, channel),
                individual=True,
                scalebar=True,
                scalebar_size=scalebar_size,
                return_rgb=True,
                downsample=downsample,
            )

            if print_histogram:
                hist_data = np.histogram(rgb[:, :, 1], bins=10, range=hist_range)
                hist = hv.Histogram(hist_data).opts(
                    color=hist_colour, alpha=0.4, toolbar=None, xlabel="Intensity"
                )

                hist_intensities = hist_data[0]
                hist_borders = hist_data[1]

                with open(os.path.join(path, channel + "_hist.png"), "w") as outfile:
                    for n in range(len(hist_intensities)):
                        outfile.write("\n")
                        outfile.write(
                            str(hist_borders[n]) + "-" + str(hist_borders[n + 1]) + ","
                        )
                        outfile.write(str(hist_intensities[n]))

                hv.save(hist, os.path.join(path, channel + "_hist.png"))

    def print_all_channels(
        self,
        max_value,
        min_value,
        scalebar_size,
        colour,
        show_legend=False,
        legend_position="bottom_left",
        print_histogram=False,
        hist_range=(0, 255),
        hist_colour="blue",
        path=".",
        median=None,
        print_nuclear_ch=False,
        nuclear_channel=None,
        nuclear_max="p99",
        nuclear_colour="blue",
        downsample=None,
    ):

        self.all_images = list()

        for index, row in self.channels.iterrows():
            if row["good"] == 1:
                channel = row["target"]
                if print_nuclear_ch:
                    ch_to_display = [
                        [channel, 0, max_value, colour, median, channel],
                        [
                            nuclear_channel,
                            0,
                            nuclear_max,
                            nuclear_colour,
                            median,
                            channel,
                        ],
                    ]
                else:
                    ch_to_display = [[channel, 0, max_value, colour, median, channel]]
                self.prepare_data(ch_to_display)
                self.combine_ch_colors()
                self.ROI_image_individual = self.rendered_image.select(
                    x=self.x_range, y=self.y_range
                )
                self.all_images.append(self.ROI_image_individual)
                rgb = self.save_image(
                    os.path.join(path, channel),
                    individual=True,
                    scalebar=True,
                    scalebar_size=scalebar_size,
                    return_rgb=True,
                    downsample=downsample,
                )

                if print_histogram:
                    hist_data = np.histogram(rgb[:, :, 1], bins=10, range=hist_range)
                    hist = hv.Histogram(hist_data).opts(
                        color=hist_colour, alpha=0.4, toolbar=None, xlabel="Intensity"
                    )
                    hist_intensities = hist_data[0]
                    hist_borders = hist_data[1]

                    with open(
                        os.path.join(path, channel + "_hist.png"), "w"
                    ) as outfile:
                        for n in range(len(hist_intensities)):
                            outfile.write("\n")
                            outfile.write(
                                str(hist_borders[n])
                                + "-"
                                + str(hist_borders[n + 1])
                                + ","
                            )
                            outfile.write(str(hist_intensities[n]))

                    hv.save(hist, os.path.join(path, channel + "_hist.png"))


@ngjit
def norm_colorize(agg, min_val, max_val, colour, tot):
    out = np.zeros((agg.shape[0], agg.shape[1], 3), dtype=np.float32)
    range_val = max_val - min_val
    col, rows = agg.shape
    for x in range(col):
        for y in range(rows):
            val = agg[x, y]
            norm = (val - min_val) / range_val
            if norm > 1:
                norm = 1

            out[x, y, 0] = tot[x, y, 0] + norm * colour[0]
            out[x, y, 1] = tot[x, y, 1] + norm * colour[1]
            out[x, y, 2] = tot[x, y, 2] + norm * colour[2]

    return out



from pathlib import Path

import numpy as np
import pytest

from imaxt_image.io.tiffimage import TiffImage
from imaxt_image.misc import bytescale


def test_bytescale():
    img = np.random.randn(100, 100)
    im = bytescale(img)
    assert im.dtype == np.uint8

    with pytest.raises(ValueError):
        bytescale(img, high=0, low=1)

    with pytest.raises(ValueError):
        bytescale(img, cmax=0, cmin=1)

    im = bytescale(img * 0)
    assert im.dtype == np.uint8

    im = bytescale(im)
    assert im.dtype == np.uint8
